#!/usr/bin/env python
# -*- coding: UTF-8 -*-

# goear-dl. Script para descargar archivos desde goear.
# Copyright (C) 2007-2013 Daniel Fuentes B.

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# ===================================================================
# ########### START EDIT HERE ###########
# ===================================================================

# Directory with the translations files (*.mo)
APP_DIR = "./po"

# ===================================================================
# ########### STOP EDIT HERE ############
# ===================================================================

# Importamos algunos módulos que luego usaremos en el script:

import gettext
import urllib
import sys
import re
import htmlentitydefs
import os.path
from os import curdir

# ===================================================================
# informacion variada:
# ===================================================================

__version__ = "0.9"
quiet_mode = False

# Definimos el idioma a utilizar en windows por gettext, ya que por
# defecto en windows no existe la variable de entorno LANG

if sys.platform.startswith("win"):
    import locale
    if os.getenv("LANG") is None:
        lang, enc = locale.getdefaultlocale()
        os.environ["LANG"] = lang
else:
    pass

# Algunas cosas para gettext (i18n / traducciones)

_ = gettext.gettext

gettext.textdomain("goear-dl")
gettext.bindtextdomain("goear-dl", APP_DIR)

# El texto que se muestra en la ayuda

texto_ayuda = _("""Usage: goear-dl.py [Options] [URL or ID]

URL Format (or ID):

Please insert a valid URL, similar to:
http://www.goear.com/listen.php?v=xxxxxxx
http://www.goear.com/listen/xxxxxxx/[song-author]
Or enter the ID of the file (has 7 number/letters)

Options:

-h, --help         Print this help text and exit
-v, --version      Print program version and exit
-i, --input FILE   Download URLs/IDs found in local FILE
    --list FILE      (write one URL/ID per line in FILE)
-q, --quiet        Activates quiet mode

Examples of usage:

** You want to download a URL. Just type one of these:
goear-dl.py http://www.goear.com/listen.php?v=xxxxxxx
goear-dl.py http://www.goear.com/listen/xxxxx/[song-author]
goear-dl.py xxxxxxx

** Download a URL/ID in quiet mode:
goear-dl.py -q http://www.goear.com/listen.php?v=xxxxx
goear-dl.py -q http://www.goear.com/listen/xxxxx/[song-author]

** Read url's from a local file and retrieve them:
goear-dl.py --list file_with_urls.txt

For more info check "README" file.""")

# ===================================================================
# Funciones y clases usadas en el programa:
# ===================================================================


class MyOpener(urllib.FancyURLopener):
    """Clase que cambia el user agent que usa python"""
    # Por defecto python usa: URLopener.version = "Python-urllib/1.xx"
    # lo cambiamos por el de IE 7
    version = "Mozilla/5.0 (Windows; U; MSIE 7.0; Windows NT 6.0; en-US)"
    # Ver http://docs.python.org/library/urllib.html#urllib._urlopener


class Goear():
    """Clase que obtiene la informacion del archivo (nombre, artista, etc)"""

    def __init__(self, audio_id):
        self.id = audio_id
        self.data = None
        # Para que funcione la clase, al incio hay que obtener la informacion
        # del archivo xml a partir de la ID (primero abrir el xml)
        url_xml = "http://www.goear.com/playersong/" + self.id
        # Cambiamos el user agent por el definido anteriormente:
        urlopen = MyOpener().open
        try:
            xmldoc = urlopen(url_xml, None)
            xml_data = xmldoc.read()
            xmldoc.close()
            # transformar el texto desde utf-8 a la codificación por defecto
            # para evita que letras raras se escriban mal en el SO usado
            data = xml_data.decode("utf-8", "ignore")
        # Mensaje en caso de error, o de no encontrarse la ruta
        except:
            data = None
            print _("[goear] Error: Failed to obtain information from Goear")
        self.data = data  # Contiene la copia de informacion del xml

    def get_id(self):
        "muestra la ID de la cancion"
        return self.id

    def url_file(self):
        "devuelve la url del archivo mp3"
        if self.data is None:
            path = None
        else:
            # Buscar la linea "track href=" que es la dirección del archivo
            # nota: nos ahorramos la busqueda, ahora todas soun una url+ID
            path = "http://www.goear.com/action/sound/get/" + self.id
        return path

    def title(self):
        "devuelve el titulo de la cancion"
        if self.data is None:
            title = ""
        else:
            # Buscar la linea "title=" que es la dirección del archivo
            inicio = self.data.find("title=")
            fin = self.data.find("\"", (inicio + 7), -1)
            title = self.data[(inicio + 7):fin]
            title = tranforma(title)
        return title

# -------------------------------------------------------------------
# Definimos la función que tranforma caracteres especiales que pueden
# aparecer en los textos (usado principalmente en el titulo/artista):


def tranforma(s):
    # se reemplaza "&amp;" por la "&" correspondiente
    s = s.replace("&amp;", "&")
    # se busca otros caracteres y simbolos usados en (x)html y xml que son
    # de la forma &***; ver http://www.w3schools.com/tags/ref_entities.asp
    matches = re.findall("&[a-zA-Z0-9]+;", s)
    hits = set(matches)
    if len(matches) > 0:
        for hit in hits:
            name = hit[1:-1]
            if name in htmlentitydefs.name2codepoint:
                s = s.replace(hit, unichr(htmlentitydefs.name2codepoint[name]))
    return s


# -------------------------------------------------------------------
# Definimos la función de descarga:


def download_file(filename, URL, quiet=False):
    """"función que descarga el archivo"""
    # Cambiamos el user agent por el definido anteriormente:
    urlretrieve = MyOpener().retrieve
    if quiet is True:
        f = urlretrieve(URL, filename)
    else:
        # Se usa my_report_hook() para mostrar el progreso en pantalla
        f = urlretrieve(URL, filename, reporthook=my_report_hook)


def my_report_hook(block_count, block_size, total_size):
    """función que muestra el progreso de la descarga de archivo"""
    # Usada si al descargar se pide mostrar el progreso (por defecto)
    total_kb = total_size / 1024
    descargado = block_count * (block_size / 1024)
    porcentaje = descargado * 100 / total_kb
    # Mostramos en el teminal la salida, se usa stdout porque no
    # inserte el alto de linea "\n" a diferencia de print
    texto_barra = _("[download] %d kb of %d kb (%3d %s)") % (descargado,
                    total_kb, porcentaje, "%")
    # Borramos la linea anterior (asi se actualiza la misma linea)
    sys.stdout.write("\b" * (len(texto_barra)))
    sys.stdout.write(texto_barra)
    sys.stdout.flush()


# -------------------------------------------------------------------
# Funcion que obtiene la ID a partir de la dirección ingresada


def get_id(audio_url):
    """"función que determina la Id de la canción de una url de goear"""
    # hay varios posibles casos:
    # a) Se ingresa directamente la ID (codigo de 7 numeros/letras)
    # b) De la forma antigua (http://www.goear.com/listen.php?v=xxxxx):
    # c) De la forma nueva (http://www.goear.com/listen/xxxxx/cancion-artista)
    if len(audio_url) == 7:
        # se ingreso la id directamente (primer caso)
        audio_id = audio_url
    elif (audio_url.find("listen.php?v=")) != -1:  # Segundo caso
        # de la forma antigua el find deberia encontrar el "listen.php?v="
        # siendo distinto de -1 (es -1 cuando no encuentra el texto pedido)
        inicio_id = audio_url.find("v=")
        # La id son 7 numeros/letras, asi que nos detenemos en el 7 caracter
        audio_id = audio_url[(inicio_id + 2):(inicio_id + 9)]
    elif (audio_url.find("listen/")) != -1:
        # la nueva forma (tercer caso)
        inicio_id = audio_url.find("listen/")
        audio_id = audio_url[(inicio_id + 7):(inicio_id + 14)]
    else:
        print _("[goear-dl] Error: Not find the address or ID for:\n %s") % (
            audio_url)
        audio_id = None
    # Regresar las ID de las URLs
    return audio_id


# -------------------------------------------------------------------
# Funcion que lee un archivo de texto del disco y devuelve su contenido


def filelist(filename):
    """lee un archivo de texto y devuelve su contenido"""
    # Se busca el filename en el directorio actual
    if os.path.exists(os.path.join(os.curdir, filename)):
        lista_lineas = []
        fichero = open(filename, "r")
        for line in fichero:
            lista_lineas.append(line.rstrip())
            # Con rstrip() Elimina \n al final de la linea
        fichero.close()
        return lista_lineas
    # Se supone que el filename es la ruta completa de la lista
    elif os.path.exists(filename):
        lista_lineas = []
        fichero = open(filename, "r")
        for line in fichero:
            lista_lineas.append(line.rstrip())
        fichero.close()
        return lista_lineas
    else:  # No se encuentra la ruta del archivo (que contiene la lista)
        print _("[goear-dl] Error: %s not found.") % (filename)
        sys.exit(1)


# -------------------------------------------------------------------
# Funcion que comprueba las opciones ingresadas, actua de acuerdo a ella
# y regresa la url de goear (si es necesaria) o termina el programa


def checkoptions():
    """Comprueba las opciones, las ejecuta y regresa las audio_url"""
    # Variables a usar en la funcion (una ya esta definida, la global)
    global quiet_mode
    opciones_ingresadas = []  # lista con las opciones
    audio_url_list = []  # lista con las URLs

# -----------------------------------------------
# Primero se comprueba la entrada de datos, hay dos posibles casos:
# a) Solo una opcion o la URL/ID, b) muchas opciones

    # Caso 1: Se ha ingresa solo 1 opcion (o url) en la linea de ordenes
    if len(sys.argv) == 2:
    # Hay tres valores posibles:
    # a) se ingreso una opcion
    # b) se ingreso la opcion de lista (sin especificar cual archivo es)
    # c) se ingreso directamente la url (o algo similar)
        if (sys.argv[1] == "-q" or sys.argv[1] == "--quiet" or
                sys.argv[1] == "-h" or sys.argv[1] == "--help" or
                sys.argv[1] == "help" or sys.argv[1] == "-v" or
                sys.argv[1] == "--version" or sys.argv[1] == "moo"):
            # como no hay url no se agrega nada a audio_url_list
            opciones_ingresadas.append(sys.argv[1])
        elif (sys.argv[1] == "--list" or sys.argv[1] == "-i" or
                sys.argv[1] == "--input"):
            # en este caso se ingreso la opcion de leer una lista desde un
            # archivo, pero NO indica el achivo que hay leer (salir con error)
            print (_("[goear-dl] Error: Please enter the list filename \
in the option: %s") % (sys.argv[1]))
            sys.exit(1)
        else:
            audio_url_list.append(sys.argv[1])

    # Caso 2: Se ingresan muchas opciones en la linea de comandos
    elif len(sys.argv) > 2:
    # Recorrer la lista de lo ingresado, se pueden tener en ella:
    # a) ordenes/opciones
    # b) opcion de lista + archivo con las URL
    # c) tener URL directamente
        ingresados = sys.argv[1:len(sys.argv)]
        m = 0
        for n in ingresados:
            m = m + 1
            if (n == "-q" or n == "--quiet" or n == "-h" or n == "--help"
                    or n == "help" or n == "-v" or n == "--version" or
                    n == "moo"):
                # Agregar a las opciones directamente
                opciones_ingresadas.append(n)
            elif (n == "--list" or n == "-i" or n == "--input"):
                # Leer el archivo y agregar el contenido a la lista de URLs
                lista_archivos = filelist(ingresados[m])
                for elementos in lista_archivos:
                    audio_url_list.append(elementos)
                # remover el archivo que contiene lista con urls de los argv
                del ingresados[m]
            else:
                audio_url_list.append(n)

# -----------------------------------------------
# Luego hay que verificar las Opciones de la linea de comandos
# y luego proceder de acuerdo a ellas. Nota: se ignora la opciones
#  --list, --imput, -i ya que se agregaron la urls en el paso anterior
    if opciones_ingresadas != []:
        for opcion in opciones_ingresadas:
            if opcion == "-h" or opcion == "--help" or opcion == "help":
                print texto_ayuda
                sys.exit()
            elif opcion == "-v" or opcion == "--version":
                print _("[goear-dl] Version: %s") % (__version__)
                sys.exit()
            elif opcion == "-q" or opcion == "--quiet":
                quiet_mode = True
            elif opcion == "moo":
                import time
                print _("No Super Cow in this program, but there is \
a dance")
                sys.stdout.write(" <(^.^<) ")
                sys.stdout.flush()
                time.sleep(2)
                sys.stdout.write("\b" * 9)
                sys.stdout.write(" <(^-^)> ")
                sys.stdout.flush()
                time.sleep(2)
                sys.stdout.write("\b" * 9)
                sys.stdout.write(" (>^.^)> ")
                sys.stdout.flush()
                time.sleep(2)
                sys.stdout.write("\b" * 9)
                sys.stdout.write(" ^(^.^)^\n")
                sys.stdout.flush()
                sys.exit()
            else:
                pass

# -----------------------------------------------
# Por ultimo comprobar que se tienen direccines (url) del audio:
# Puede darse casos como el que sigue: goear.py -q -q (bien raro)
# en donde audio_url_list estaria vacia (preguntar en ese caso).
    if audio_url_list == []:
        audio_url_list.append(str(raw_input(_("[goear-dl] you must \
provide at least one URL or ID:\n"))))

# -----------------------------------------------
# Finalmente la funcion devuelve la lista de las ID (o las URL) de cada
# cancion, pero no comprueba si son validas (comprobar con funcion get_id())
    return audio_url_list


#====================================================================
# Iniciamos el programa
#====================================================================

if __name__ == "__main__":
    songs_list = []
    # Comprobar si se ingreso algun dato en la linea de ordenes.Si no
    # se ingresa nada sys.argv solo tendra 1 dato (nombre del script)
    if len(sys.argv) == 1:
        # No hay ordenes o url, por lo que hay que pedir la dirección
        songs_list.append(str(raw_input(_("[goear-dl] you must \
provide at least one URL or ID:\n"))))
    else:
        songs_list = checkoptions()  # funcion que devuelve las ID

    # Se revisa la lista con las url de goear, para cada una obtiene la
    # ID de cada cancion y luego se procede con la descarga:
    for audio_url in songs_list:
        audio_id = get_id(audio_url)
        try:
            # intenta obtener la informacion del archivo
            cancion = Goear(audio_id)
        except:
            pass
        # Comprobar la direccion e iniciar la descarga del archivo
        if audio_id is None:
            # se pasa a la siguiente en la lista debido a que la URL
            # ingresada no tiene el formato esperado (ver funcion get_id())
            pass
        elif cancion.url_file() is None:
            # Si la ID no se encuentra (no se encuentra informacion en goear)
            # el cancion.url_file() es None, por lo que pasamos:
            print _("[goear] Error: %s not foud") % (audio_id)
            pass
        else:
           # En este caso si se puede iniciar la descarga
            print _("[goear] Extracting audio information for: %s") % (
                audio_id)
            # Primero Creamos el nombre del archivo
            nombre_archivo = (cancion.get_id() + " - " + cancion.title() +
                            ".mp3")
            # Imprimir informacion del nombre del archivo si es necesario
            if quiet_mode is False:
                print _("[download] Destination: \"%s\"") % (
                    nombre_archivo)
            # Intentar iniciar la descarga
            try:
                # download_file() muestra el progreso
                download_file(nombre_archivo, cancion.url_file(), quiet_mode)
                if quiet_mode is False:
                    # Salto de linea la final de la barra de descarga
                    # Es necesario ya que sys.stdout.write() no inserta el
                    # salto de linea \n al final de la barra de progreso)
                    sys.stdout.write("\n")
                    sys.stdout.flush()
                # Mensaje avisando de que se termino la descarga
                print _("[download] Done. Open the file: \"%s\"") % (
                    nombre_archivo)
            except:
                # No se puede abrir/descargar, se pasa a la siguiente
                print _("[download] Error: Failed to save: \"%s\"") % (
                    nombre_archivo)
                pass
